/*
 * Copyright (c) 2011, Noel Lemouel
 * All rights reserved.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include "serial.h"

#define VERSION __DATE__

/* needs to be accessed outside of main after CTRL+C */
static FILE* fd0 = NULL;
static int fdT = (int)NULL;
static unsigned char* bin = NULL;

static int sc_open(FILE**, const char*, struct stat*);

void sig_handler(int signum)
{
	// printf("signal %d\n", signum);

	switch(signum){
		case(SIGINT):
			printf("\n\n");
			break;
		case(SIGALRM):
			printf("\ntimeout\n\n");
			break;
		default:
			printf("unexpected signal [%d]\n", signum);
			break;
	}

	if(bin) free(bin);
	if(fd0) fclose(fd0);
	if(fdT) tio_exit();
	serial_close();

	exit(0);
}

int download(unsigned int base, unsigned int size, unsigned char* bin, unsigned char type)
{
	int i, j, k;
	unsigned char rc[4];
	unsigned char buf[128];

	/* send the command */

	// | D | L | a3 | a2 | a1 | a0 | - | s3 | s2 | s1 | s0 | - | - | - | - | DT |
	// DT = 'C' => code
	// DT = 'D' => data

	buf[0] = 'D';
	buf[1] = 'L';
	buf[2] = (base & 0xff000000) >> 24;
	buf[3] = (base & 0x00ff0000) >> 16;
	buf[4] = (base & 0x0000ff00) >> 8;
	buf[5] = base & 0x000000ff;
	buf[6] = 0xaa;
	buf[7] = (size & 0xff000000) >> 24;
	buf[8] = (size & 0x00ff0000) >> 16;
	buf[9] = (size & 0x0000ff00) >> 8;
	buf[10] = size & 0x000000ff;
	buf[11] = 0xaa;
	buf[12] = 0xaa;
	buf[13] = 0xaa;
	buf[14] = 0xaa;
	buf[15] = type;

	serial_tx(buf, 16);

	/* read the answer */

	alarm(3);

	k = 0;
	for(;;)
	{
		j = serial_rx(buf, 128);
		if(j)
		{
			i = 0;
			while(j--) rc[k] = buf[i++], k++;
		}
		if(k >= 4) break;
		usleep(10000);
	}

	printf("dl |0x%02x|0x%02x|0x%02x|0x%02x|\n", rc[0], rc[1], rc[2], rc[3]);

	if(strncmp("ACK", (char*)rc, 4))
	{
		printf("\ninvalid response\n");
		sig_handler(SIGINT);
	}

	alarm(0);

	printf(" \n");

	/* verify ACK... */

	printf("uploading data... ");

	/* upload data */

	serial_tx(bin, size);

	printf("done");

	return 0;
}

/* expects 0xdeadbeef or deadbeef */
unsigned int atox(unsigned char *ip_str)
{
	unsigned int op_val=0, i=0, ip_len = strlen((char*)ip_str);

	if(strncmp((char*)ip_str, "0x", 2) == 0){
		ip_str +=2;
		ip_len -=2;
	}

	for(i=0;i<ip_len;i++){
		op_val *= 0x10;
		switch(ip_str[i]){
			case 'A':
			case 'a':
				op_val += 0xa;
				break;
			case 'B':
			case 'b':
				op_val += 0xb;
				break;
			case 'C':
			case 'c':
				op_val += 0xc;
				break;
			case 'D':
			case 'd':
				op_val += 0xd;
				break;
			case 'E':
			case 'e':
				op_val += 0xe;
				break;
			case 'F':
			case 'f':
				op_val += 0xf;
				break;
			case '0' ... '9':
				op_val += (ip_str[i] - 0x30);
				break;
			default :
				op_val += 0x0;
				break;
		}
	}

	return op_val;
}

int main(int argc, char* argv[])
{
	int i, j, n;
	int term;
	int baudrate;
	char *tty;
	unsigned char c;
	unsigned char rx[512];
	struct stat info;

	/* check arguments */

	term = 0;			/* no terminal */
	baudrate = 9600; 	/* default baudrate */

   if((argc == 2) && !strcmp("-v", argv[1]))
   {
      printf("UNIX %s\n", VERSION);
      return 0;
   }

	if((argc >= 5) && (argc % 2)){

		if( sc_open(&fd0, argv[2], &info) ) return -1;

		if( strcmp("libftdi", argv[1]) )
		{
			printf("DEV ignored [libftdi version]\n");
			return -1;
		}

		for(i=5; i<argc-1; i++)		/* last argv can't be an opt */
		{
			if( !strcmp("-t", argv[i]) )
			{
				tty = argv[i+1]; term = 1; // printf("term = 1\n");
			}
			if( !strcmp("-b", argv[i]) )
			{
				baudrate = atoi(argv[i+1]); // printf("baudrate = %d\n", baudrate);
			}
		}
	}else{
		printf("usage: %s DEV *.bin 0xXXXXXXXX 0xXX [-b baudrate] [-t tty]\n", argv[0]);
		return -1;
	}

	printf("\n-- ORX serial loader --\n\n");

	/* install signals handler */

	signal(SIGINT, sig_handler);
	signal(SIGALRM, sig_handler);

	/* load bin file */

	bin = malloc(info.st_size);
	if(!bin){
		printf("error: [alloc. %d] failed\n", (int)info.st_size);
		fclose(fd0);
		return -1;
	}else{
		// printf("[alloc. %d]\n", (int)info.st_size);
	}

	fread(bin, 1, info.st_size, fd0);
	fclose(fd0);
	fd0 = NULL;

	/* ... */

	if(serial_open(baudrate)) return -1;

	/* ... */

	if(term) printf("tty    : %s\n", tty);
	printf("port   : %s [%d]\n", argv[1], baudrate);
	printf("type   : 0x%02x [ignored]\n", atox((unsigned char*)argv[4]));
	printf("size   : 0x%08x\n", (int)info.st_size);
	printf("address: 0x%08x\n", atox((unsigned char*)argv[3]));

	download(atox((unsigned char*)argv[3]), info.st_size, bin, 'C');

	/* ... */

	if(term)
	{
		printf("\n\n[terminal mode] CTRL+D to quit\n\n");

		fdT = tio_init(tty);

		for(;;)
		{
			while(1)
			{

				n = io_read(&c);
				if(n != -1){
					break;
				}else if(errno = EAGAIN){
					/* ... */
				}

				i = serial_rx(rx, 512);
				if(i)
				{
					for(j=0; j<i; j++)
					{
						tio_putchar(rx[j]);
					}
				}

				usleep(10000);
			}

			// x_printf("*break*");

			if(c == 0x04)	/* ^D */
			{
				sig_handler(SIGINT);
			}else{
				serial_tx(&c, 1);
				// tio_putchar(c);
			}
		}
	}

	sig_handler(SIGINT);

	return 0;
}

static int sc_open(FILE** fd, const char* name, struct stat* info)
{
	if (stat(name, info) == -1)
	{
		if (errno == ENOENT)
		{
			(void) fprintf( stderr, "error: '%s' does not exist\r\n", name );
		}else{
			(void) fprintf( stderr, "error: stat() fails for '%s'\r\n", name );
		}
		return -1;
	}

	/* checking S_ISLNK(m) is not required since 'stat' follows sym. links */
	if (!S_ISREG(info->st_mode))
	{
		(void) fprintf( stderr, "error: '%s' is a not a regular file (or a sym. link)\r\n", name);
		return -1;
	}

	if ((*fd = fopen(name, "r")) == NULL)
	{
		(void) fprintf( stderr, "error: '%s' open failed\r\n", name );
		return -1;
	}

	/* printf("size: %d\n", (int)info->st_size); */

	return 0;   /* no error */
}

